
#include "config.h"

/*
#ifdef HAVE_CSTRING
#include <cstring>
#else
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#endif
*/

#ifdef HAVE_CMATH
#include <cmath>
#else
#include <math.h>
#endif

#include <algorithm>
#include <limits>

#include "Bbox2D.h"

#if 0
/*!*************************************
* Create a boundary box given the lower left and upper right corner.
* @param ptll The coordinates of the lower left corner.
* @param ptur coordinates of the upper right corner.
*/
Bbox2D::Bbox2D(const Point2D &ptll , const Point2D & ptur)
{
	limits[0] = ptll ;
	limits[1] = ptur ;
} /* ctor */
#else
/*!*************************************
* @brief Create a boundary box given two corner points.
* @param pt1 The first corner.
* @param pt2 The second corner.
*/
Bbox2D::Bbox2D(const Point2D & pt1 , const Point2D & pt2)
{
	limits[0] = limits[1] = pt1 ;
	include (pt2) ;
} /* ctor */
#endif

/*!*************************************
* Create an invalid bounding box.
* Default ctor.
*/
Bbox2D::Bbox2D()
{
	/* set both edges to a non-overlapping region.
	limits[0].coord[1] = limits[0].coord[0] = numeric_limits<double>max() ;
	limits[1].coord[1] = limits[1].coord[0] = numeric_limits<double>min() ;
	*/
	limits[0].coord[1] = limits[0].coord[0] 
	= limits[1].coord[1] = limits[1].coord[0] 
		= numeric_limits<double>::quiet_NaN() ;
} /* ctor */

/*!*************************************
* Extend the bounding box such that a point is inside.
* @param[in] pt The point that may expand the bounding box.
*/
void Bbox2D::include(const Point2D &pt)
{
	if ( isnan(limits[0].coord[0]) )
		/* if undefined until here: bouding box is that
		* single point.
		*/
		limits[0] = limits[1] = pt ;
	else
	{
		/* if there are already points inside: extend
		*/
		limits[1].coord[0] = max( limits[1].coord[0], pt.coord[0]) ;
		limits[1].coord[1] = max( limits[1].coord[1], pt.coord[1]) ;
		limits[0].coord[0] = min( limits[0].coord[0], pt.coord[0]) ;
		limits[0].coord[1] = min( limits[0].coord[1], pt.coord[1]) ;
	}
} /* include */

/*!********************************************
* Decide whether a point is inside (or on the rim) of the bounding box.
* @param[in] pt The point to be tested for insidedness.
* @return True if the point is inside.
*/
bool Bbox2D::isInside(const Point2D & pt) const
{
	if ( isnan(limits[0].coord[0]) )
		/* if undefined until here: no point is inside.
		* single point.
		*/
		return false ;
	else
	{
		if ( pt.coord[0] >= limits[0].coord[0] && 
			pt.coord[0] <= limits[1].coord[0] && 
			pt.coord[1] >= limits[0].coord[1] && 
			pt.coord[1] <= limits[1].coord[1])
			return true;
		else
			return false;
	}
} /* isInside */

/*!***********************************************
* Compute the lower right corner of the bounding box.
* @return The point in the bounding box with maximum x and minimum y coordinate.
*/
Point2D Bbox2D::lowerRight() const
{
	return Point2D( limits[1].coord[0], limits[0].coord[1]) ;
}

/*!***********************************************
* Compute the upper left corner of the bounding box.
* @return The point in the bounding box with minimum x and maximum y coordinate.
*/
Point2D Bbox2D::upperLeft() const
{
	return Point2D( limits[0].coord[0], limits[1].coord[1]) ;
}

/*!****************************************
* @brief Compute the bounding box of the quadrangle that is rotated-translated.
* @param[in] rt The rotation-translation to be applied
* @return The bounding box which includes all 4 rotated corner points of
*   this bounding box.
*/
Bbox2D Bbox2D::apply(const RotTrans2D & rt) const
{
	if ( isnan(limits[0].coord[0]) )
		return Bbox2D() ;

        /* Compute the coordinates of each of the old corners
        * under the tranformation.
        */
        Point2D imgCorn[4] ;
	imgCorn[0] = limits[0].apply(rt) ;
	imgCorn[1] = lowerRight().apply(rt) ;
	imgCorn[2] = limits[1].apply(rt) ;
	imgCorn[3] = upperLeft().apply(rt) ;

        /* Build the bounding box of these 4 images (which is the
        * bounding box of the result.
        */
        Bbox2D img(imgCorn[0],imgCorn[2]) ;
        img.include(imgCorn[1]) ;
        img.include(imgCorn[3]) ;

	return img ;
} /* apply */
