/*
* MPIA - GEIRS project
*
* "@(#) $Id$"
*
*/

#include "config.h"

#include <string>
#include <cstring>
#include <ctime>
#include <cstdlib>
#include <fstream>
#include <limits>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "FitsImg2Asc.h"
#include "Window.h"
#include "Histos.h"

using namespace std ;
using namespace CCfits ;

/** Constructor.
* @param[in] fitsname The name of the FITS file to be read.
* @since 2012-11-25
* @author Richard J. Mathar
*/
FitsImg2Asc::FitsImg2Asc(char *fitsIname) : iname()
{
	// cout << __FILE__ << " " << __LINE__ << *fitsIname << endl ;
	iname.push_back(fitsIname) ;
} /* Fits2Img2Asc::Fits2Img2Asc */

/** Constructor.
* @param[in] fitsname The names of the FITS file to be read.
* @since 2013-01-29
* @author Richard J. Mathar
*/
FitsImg2Asc::FitsImg2Asc(int argc,char *fitsIname[]) : iname()
{
	for(int f=0 ; f < argc ; f++)
	{
		// cout << __FILE__ << " " << __LINE__ << *fitsIname << endl ;
		iname.push_back(fitsIname[f]) ;
	}
} /* Fits2Img2Asc::Fits2Img2Asc */

/** Destructor.
*/
FitsImg2Asc::~FitsImg2Asc()
{
} /* FitsImg2Asc::dtor */

/** Dump the entire image in ASCII style to cout.
*/
void FitsImg2Asc::dump()
{
	FITS * ifits = new FITS(iname[0], CCfits::Read) ;
#ifdef DEBUG
	FITS::setVerboseMode(true) ;
#endif
	PHDU & iex = ifits->pHDU() ;
	iex.readAllKeys() ;

	int datasec[4] ;

	const int naxes = iex.axes() ;

	/* the size along x, alont y (both pixels) and z (if cubes)
	*/
	long axs[3] = {0L,0L,1L} ;
        for(int i=0 ; i < naxes ; i++)
                axs[i] = iex.axis(i) ;

	datasec[0]=0 ;
	datasec[1]= axs[0] ;
	datasec[2]=0 ;
	datasec[3]= axs[1] ;
	dump(datasec, ifits) ;
	/* remove the allocated resoures
	*/
	delete ifits ;
	ifits =0 ;
} /* FitsImg2Asc::dump */

/** Dump the image in ASCII style to cout.
* @param[in] range Four values denoting the lower value of x, the upper value
*  of x (exclusive), the lower value of y and the upper value of y (exclusive)
*  of the pixel region to be plottet.
*/
void FitsImg2Asc::dump(string range)
{
	int r[4] ;
	Window::fourPix(range,r,false) ;
	FITS * ifits = new FITS(iname[0], CCfits::Read) ;
#ifdef DEBUG
	FITS::setVerboseMode(true) ;
#endif
	dump(r,ifits) ;
	/* remove the allocated resoures 
	*/
	delete ifits ;
	ifits =0 ;
}

/** Dump the image in ASCII style to cout.
* @param[in] range Four values denoting the pixel range.
*  The lower value of x, the upper value
*  of x (exclusive), the lower value of y and the upper value of y (exclusive)
*  of the pixel region to be plotted.
*/
void FitsImg2Asc::dump(int range[4], FITS * ifits)
{
	PHDU & iex = ifits->pHDU() ;
	iex.readAllKeys() ;

	int datasec[4] ;

	const int naxes = iex.axes() ;

	/* the size along x, along y (both pixels) and z (if cubes)
	*/
	int axs[3] = {0L,0L,1L} ;
        for(int i=0 ; i < naxes ; i++)
                axs[i] = (int)(iex.axis(i)) ;

#if 0
	/* todo: range checks
	if (range[0] < 0 || range[0] >= range[1] )
		range[0] = 0 ;
	*/
#endif

	map<String, Keyword*> & allkw = iex.keyWord() ;

	map<String, Keyword*>::iterator kitr = allkw.begin() ;
	for ( ; kitr != allkw.end() ; kitr++)
	{
		const string & knam = (*kitr).first ;
		const Keyword * thisk = (*kitr).second ;
		if ( knam == "SAVEAREA" || knam == "DATASEC")
		{
			string tsiz ;
			thisk->value(tsiz) ;
			Window::fourPix(tsiz,datasec,false) ;
		}
	}

	time_t now = time(NULL) ;
	cout << "# created by "__FILE__ << " " << ctime(&now) << endl ;
	cout << "# from " << iname[0] << " ranges [" << range[0] << ":" << range[1] 
		<< "," << range[2] << ":" << range[3] << "]" << endl ;

	/* treat all data as float values, independent of BITPIX
	*/
	valarray<float> farr ;
	iex.read(farr,1,axs[0]*axs[1]*axs[2]) ;

	/* put all values to the cout, independent whether they match any BLANK
	* value that might havae been defined in the FITS header
	*/
	for(int y= max(0,range[2]) ; y < min(axs[1],range[3]) ; y++)
	{
		for(int x= max(0,range[0]) ; x < min(axs[0],range[1]) ; x++)
			cout << x+datasec[0] << " " << y+datasec[2] << " " << farr[x+axs[0]*y] << endl ;
		cout << endl ;
	}

} /* FitsImg2Asc::dump */

/** Generate a gnuplot X11 window with the histogram(s).
* @param[in] Nbin The number of bins in the histogram.
*   A value less than zero will trigger default setting of Nbin by the program.
*   The actual number will then be the square root of the maximum number
*   of data values in any of the arrays.
* @param[in] range The lower [0] and upper [1] limit of the points on the
*   horizontal axis (ADU) to be used.
* @param[in] detsize A string of the form [xmin:xmax,ymin:ymax] selecting
*   a XY ranges of pixels. If empy, the entire frame is examined.
* @param[in] doLogHist Print a logarithmic, not a linear vertical scale.
* @param[in] epsout If not an empty string, use this file for plotting, not the X11 screen.
* @param[in] txtout If not an empty string, use this text file to store the histogram, not the X11 screen.
* @since 2013-01-29
*/
void FitsImg2Asc::histo(int Nbin, double range[2], string detsize, const bool doLogHist, string epsout, string txtout)
{
	/* treat all data as float values, independent of BITPIX
	*/
	vector< valarray<float> > arr( iname.size()) ;

	string contName ;

	/* read all pixels of all files
	*/
	for(int f=0 ; f < iname.size() ; f++)
	{
		FITS * ifits = new FITS(iname[f], CCfits::Read) ;
#ifdef DEBUG
		FITS::setVerboseMode(true) ;
#endif

		int naxes ;
		/* the size along x, alont y (both pixels) and z (if cubes)
		*/
		long axs[3] = {0L,0L,1L} ;

		int hduidx = 0;
		try
		{
			ExtHDU & cex = ifits->currentExtension() ;
			/* reading the index of a primary HDU throws a 
			* Cannot read HDU in FITS file: No HDU with index 0
			* exception.
			*/
			hduidx = cex.index() ;
			naxes = cex.axes() ;
		        for(int i=0 ; i < naxes ; i++)
		                axs[i] = cex.axis(i) ;
		}
		catch (FITS::NoSuchHDU)
		{
			PHDU & pex = ifits->pHDU() ;
			naxes = pex.axes() ;
		        for(int i=0 ; i < naxes ; i++)
		                axs[i] = pex.axis(i) ;
		}

		/* [0] = xstart, [1]=xend (exclusive), [2]=ystart, [3]=yend (exclusive)
		* in all four cases 0-based.
		*/
		int datasec[4] ;
		if ( detsize.length() == 0 )
		{
			datasec[0] = datasec[2] = 0;
			datasec[1] = (int)axs[0] ;
			datasec[3] = (int)axs[1] ;
			if ( hduidx > 0 )
			{
				ExtHDU & iex = ifits->currentExtension() ;
				iex.read(arr[f],1L,axs[0]*axs[1]*axs[2]) ;
			}
			else
			{
				PHDU & iex = ifits->pHDU() ;
				iex.read(arr[f],1L,axs[0]*axs[1]*axs[2]) ;
			}
		}
		else
		{
			/* extract pixel regions, 1-based
			*/
			Window::fourPix(detsize,datasec,false) ;
			/* move on to zero-based. datasec[1,3] stay as the are,
			* because they are reintepreted as non-inclusive then.
			*/
			datasec[0]-- ;
			datasec[2]-- ;
			/* correct user input that has been too wide
			*/
			datasec[0] = max(datasec[0],0) ;
			datasec[0] = min(datasec[0],(int)axs[0]) ;
			datasec[1] = max(datasec[1],0) ;
			datasec[1] = min(datasec[1],(int)axs[0]) ;
			datasec[2] = max(datasec[2],0) ;
			datasec[2] = min(datasec[2],(int)axs[1]) ;
			datasec[3] = max(datasec[3],0) ;
			datasec[3] = min(datasec[3],(int)axs[1]) ;
			/* read into a temporary full-frame array
			*/
			valarray<float> fullf ;
			if ( hduidx > 0 )
			{
				ExtHDU & iex = ifits->currentExtension() ;
				iex.read(fullf,1,axs[0]*axs[1]*axs[2]) ;
			}
			else
			{
				PHDU & iex = ifits->pHDU() ;
				iex.read(fullf,1,axs[0]*axs[1]*axs[2]) ;
			}

			int winx = datasec[1]-datasec[0] ;
			int winy = datasec[3]-datasec[2] ;
#if 0
			/* this does not work with a bug in g++ 4.3
			*/
			arr[f] = valarray<float>(winx*winy) ;
#else
			arr[f] = valarray<float>() ;
			arr[f].resize(winx*winy) ;
#endif
			for(int y=0 ; y < winy ; y++)
			for(int x=0 ; x < winx ; x++)
				arr[f][x+y*winx] = fullf[x+datasec[0]+(y+datasec[2])*axs[0]] ;
		}

		if ( f== 0 )
		{
			/* try to grab some canconical name 
			*/
			Window win(iname[f],2048,2048) ;
			contName=win.canonName() ;
		}

		/* remove the allocated resoures with the ctor of ifits within the ctor
		*/
		delete ifits ;
		ifits =0 ;
	} /* end of loop over all FITS files */

	/* generate histogram(s)
	*/
	Histos hs(arr,Nbin,iname,contName,range) ;

	/* Create a temporary data file in /tmp
	*/
	int gplfileN = (txtout.size() > 0) ? txtout.size() : 22 ;
	char * gplDfile = new char[gplfileN+1] ;
	if ( txtout.size() > 0 )
	{
		strncpy(gplDfile,txtout.c_str(),txtout.size()+1) ;
	}
	else
	{
		strncpy(gplDfile,"/tmp/fitsImg2AscXXXXXX",gplfileN+1) ;
		int fd  = mkstemp(gplDfile) ;
		close(fd) ;
	}
	hs.dumpFil(gplDfile) ;

	/* Create a temporary gnuplot command file in /tmp
	*/
	gplfileN = 22 ;
	char * gplfile = new char[gplfileN+1] ;
	strncpy(gplfile,"/tmp/fitsImg2AscXXXXXX",gplfileN+1) ;
	int fd  = mkstemp(gplfile) ;
	close(fd) ;
	hs.gnuplot(gplDfile,gplfile,doLogHist,epsout) ;

	string cmd = "gnuplot -title 'gnuplot " + string(gplfile) + "' -persist " + string(gplfile) ;
	system(cmd.c_str()) ;

	if ( txtout.size() == 0)
		unlink(gplDfile) ;

	delete[] gplDfile ;

	unlink(gplfile) ;
	delete[] gplfile ;
} /* FitsImg2Asc::histo */

/** Generate a FITS file with a bad mask of pixels.
* @param[in] range The lower [0] and upper [1] limit of the points on the
*   horizontal axis (ADU) to be used.
* @param[in] detsize A string of the form [xmin:xmax,ymin:ymax] selecting
*   a XY ranges of pixels. If empy, the entire frame is examined.
* @param[in] fitsout Name of the FITS output file.
* @param[in] flipxy If the LSB bit (bit 0) is set, flip along x coordinate.
*   If the  bit  1 is set, flip along y coordinate.
* @param[in] txtout Name of the file with the GEIRS specific list of bad pixels.
*   Not used if null or an empty file name.
* @since 2013-12-04
*/
void FitsImg2Asc::badMask(double range[2], string detsize, string fitsout, int flipxy, string txtout, bool iraf)
{
	/* treat all data as float values, independent of BITPIX
	*/
	vector< valarray<float> > arr( iname.size()) ;

	string contName ;

	/* read all pixels of the first file
	*/
	for(int f=0 ; f < iname.size() ; f++)
	{
		FITS * ifits = new FITS(iname[f], CCfits::Read) ;
#ifdef DEBUG
		FITS::setVerboseMode(true) ;
#endif

		int naxes ;
		/* the size along x, alont y (both pixels) and z (if cubes)
		*/
		long axs[3] = {0L,0L,1L} ;

		int hduidx = 0;
		try
		{
			ExtHDU & cex = ifits->currentExtension() ;
			/* reading the index of a primary HDU throws a 
			* Cannot read HDU in FITS file: No HDU with index 0
			* exception.
			*/
			hduidx = cex.index() ;
			naxes = cex.axes() ;
		        for(int i=0 ; i < naxes ; i++)
		                axs[i] = cex.axis(i) ;
		}
		catch (FITS::NoSuchHDU)
		{
			PHDU & pex = ifits->pHDU() ;
			naxes = pex.axes() ;
		        for(int i=0 ; i < naxes ; i++)
		                axs[i] = pex.axis(i) ;
		}

		/* Input data section [0] = xstart, [1]=xend (exclusive), [2]=ystart, [3]=yend (exclusive)
		* in all four cases 0-based.
		*/
		int datasec[4] ;

		int winx,winy ;

		if ( detsize.length() == 0 )
		{
			datasec[0] = datasec[2] = 0;
			datasec[1] = (int)axs[0] ;
			datasec[3] = (int)axs[1] ;
			if ( hduidx > 0 )
			{
				ExtHDU & iex = ifits->currentExtension() ;
				iex.read(arr[f],1L,axs[0]*axs[1]*axs[2]) ;
			}
			else
			{
				PHDU & iex = ifits->pHDU() ;
				iex.read(arr[f],1L,axs[0]*axs[1]*axs[2]) ;
			}
			winx = datasec[1]-datasec[0] ;
			winy = datasec[3]-datasec[2] ;
		}
		else
		{
			/* extract pixel regions, 1-based
			*/
			Window::fourPix(detsize,datasec,false) ;
			/* move on to zero-based. datasec[1,3] stay as the are,
			* because they are reintepreted as non-inclusive then.
			*/
			datasec[0]-- ;
			datasec[2]-- ;
			/* correct user input that has been too wide
			*/
			datasec[0] = max(datasec[0],0) ;
			datasec[0] = min(datasec[0],(int)axs[0]) ;
			datasec[1] = max(datasec[1],0) ;
			datasec[1] = min(datasec[1],(int)axs[0]) ;
			datasec[2] = max(datasec[2],0) ;
			datasec[2] = min(datasec[2],(int)axs[1]) ;
			datasec[3] = max(datasec[3],0) ;
			datasec[3] = min(datasec[3],(int)axs[1]) ;
			/* read into a temporary full-frame array
			*/
			valarray<float> fullf ;
			if ( hduidx > 0 )
			{
				ExtHDU & iex = ifits->currentExtension() ;
				iex.read(fullf,1,axs[0]*axs[1]*axs[2]) ;
			}
			else
			{
				PHDU & iex = ifits->pHDU() ;
				iex.read(fullf,1,axs[0]*axs[1]*axs[2]) ;
			}

			winx = datasec[1]-datasec[0] ;
			winy = datasec[3]-datasec[2] ;
#if 0
			/* this does not work with a bug in g++ 4.3
			*/
			arr[f] = valarray<float>(winx*winy) ;
#else
			arr[f] = valarray<float>() ;
			arr[f].resize(winx*winy) ;
#endif
			for(int y=0 ; y < winy ; y++)
			for(int x=0 ; x < winx ; x++)
				arr[f][x+y*winx] = fullf[x+datasec[0]+(y+datasec[2])*axs[0]] ;
		}

		FITS * ofits ;
		if ( iraf)
			ofits = new FITS(fitsout, LONG_IMG,naxes,axs) ;
		else
			FITS * ofits = new FITS(fitsout, FLOAT_IMG,naxes,axs) ;
		PHDU & phd = ofits->pHDU() ;

		/* create an array of axs[0] X axs[1] size and initialize with NaN
		*/
		const long npx = axs[0]*axs[1] ;
		valarray<float> bdata(npx) ;
		valarray<int> idata(npx) ;
		for (int p=0 ; p < npx ; p++)
		{
			idata[p] = 1 ;
			bdata[p] = numeric_limits<float>::quiet_NaN() ;
		}

		int badcnt =0 ;
		for(int y=0 ; y < winy ; y++)
		for(int x=0 ; x < winx ; x++)
		{
			const float thisv = arr[f][x+y*winx] ;
			int targx =x+datasec[0] ;
			int targy =y+datasec[2] ;

			if ( flipxy & 1)
			{
				/* flip if requested x (right-left) */
				targx = axs[0]-1-targx ;
			}
			if ( flipxy & 2)
			{
				/* flip if requested y (up-down) */
				targy = axs[1]-1-targy ;
			}

			if ( thisv >= range[0] && thisv <= range[1])
			{
				idata[targx+targy*axs[0]] = 0 ;
				bdata[targx+targy*axs[0]] = 1.0 ;
			}
			else
				badcnt++ ;
		}

		if ( txtout.size() > 0 )
		{
			ofstream otxt(txtout.c_str()) ;
			if ( !otxt)
				cerr << "Cannot open " << txtout << endl ;
			else
			{
				/* 1 based GEIRS specific text file (x and y)
				*/
				for (int p=0 ; p < npx ; p++)
				{
					if ( isnan(bdata[p]) )
					{
						int y = p / axs[0] ;
						int x = p % axs[0] ;
						otxt << (1+x) << " " << (1+y) << endl ;
					}
				}
			}
		}

		if ( iraf)
			phd.write(1L,npx,idata) ;
		else
			phd.write(1L,npx,bdata) ;
		phd.writeDate() ;
		/* set some remarks in the comments... date ... */
		phd.addKey("ORIGFILE",iname[f],"Name of flat field input file") ;
		phd.addKey("MINCUT",range[0],"[adu] lower cut of pix admission") ;
		phd.addKey("MAXCUT",range[1],"[adu] upper cut of pix admission") ;
		phd.addKey("DATASEC",detsize,"Pixel range sampled") ;
		phd.addKey("NBAD",badcnt,"[ct] Number of bad pixels in sample range") ;


		/* remove the allocated resoures with the ctor of ifits within the ctor
		*/
		delete ifits ;
		delete ofits ;
		ifits =0 ;
		ofits =0 ;
	} /* end of loop over all FITS files */

} /* FitsImg2Asc::badmask */
