#include "Line2D.h"

/*!*************************************
* Create a line that connects two given points.
* @param strt The starting point.
* @param fini The terminal point.
*/
Line2D::Line2D(const Point2D  &strt, const Point2D & fini)
{
	pts[0] = strt ;
	pts[1] = fini ;
} /* ctor */

/*!*************************************
* Create a line of zero length at the origin.
*/
Line2D::Line2D()
{
	/* default ctor for both points... */
} /* default ctor */

/*!*************************************
* Determine the point at line parameter t.
* @param t The line parameter.
* @return A point at distance t relative to the starting point.
*  The point is the starting point if t=0 and the terminal point if t=1.
*/
Point2D Line2D::at(double t) const
{
	/* generate the point at strt+t*(fini-strt) = (1-t)*strt+t*fini.
	*/
	return Point2D( (1.-t)*pts[0].coord[0]+t*pts[1].coord[0],
			(1.-t)*pts[0].coord[1]+t*pts[1].coord[1]
		) ;
} /* at*/

/*!*************************************
* Create a vector (i.e, a direction) with
* an orientation turned 90 degrees ccw.
*/
Point2D Line2D::turn90() const
{
	/* directional vector of this line
	*/
	Point2D dire = pts[0].to(pts[1]) ;
	return dire.turn90() ;
} /* turn90 */

/*!**************************************
* Decide whether a point is to the left or to the right
* of the line (the line considered of infinite length)
* @param pt The pivotal point.
* @return True if the point to the left or on the line.
*/
bool Line2D::leftFrom(const Point2D & pt)  const
{
	/* Decompose pt = strt+alpha*(fini-start)+beta*orth,
	* where orth is a vector orthogonal to the line. Build dot products
	* pt . orth = strt.orth + beta *orth^2.
	* beta = (pt-strt).orth/ orth^2 ;
	* The sign of beta determines the result, and the sign of orth^2 is positve.
	*/
	const Point2D orth = turn90() ;
	const double beta = pts[0].to(pt).dotprod(orth) ;
	if ( beta >=0.)
		return true;
	else
		return false;
} /* leftFrom */

/*!**************************************
* Determine the point of intersection between this infinite
* line and another infinite line.
* @param oth The other line (virtually  extended both ways to infinity)
* @return The point of intersection.
*/
Point2D Line2D::intersectInf(const Line2D & oth)  const
{
	/* Write pt = strt+alpha*(fini-start) = oth.strt +beta*(oth.fini-oth.start)
	* with two line parameters alpha and beta along both lines.
	* alpha*(fini-start)-beta*(oth.fini-oth.start) = oth.strt-strt.
	* alpha*(fini-start)+beta*(oth.start-oth.fini) = oth.strt-strt.
	* which is a 2x2 linear system of equations if decomposed
	* in x and y components.
	*/
	double M[2][2] ;
	double rhs[2] ;
	for(int r=0 ; r <2 ; r++)
	{
		M[r][0] = pts[1].coord[r] - pts[0].coord[r] ;
		M[r][1] = oth.pts[0].coord[r] - oth.pts[1].coord[r] ;
		rhs[r] = oth.pts[0].coord[r] - pts[0].coord[r] ;
	}
	/* determinant of the linear system of equations... could
	* be checked for values near zero to detect almost parallel (degenerate) cases
	*/
	const double det = M[0][0]*M[1][1] - M[0][1]*M[1][0] ;
	/* Cramer's rule: insert the right hand side into the matrix
	* and re-compute the determinant. The parameter beta is not
	* computed explicitly.
	*/
	const double alpha = (rhs[0]*M[1][1] - rhs[1]*M[0][1])/det ;

	return at(alpha) ;
} /* intersectInf */
