
#include "config.h"

#ifdef HAVE_CSTRING
#include <cstring>
#else
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#endif
#include <cmath>

#include "Point2D.h"

/* because we needed to use a forward reference in Point2D.h,
* we unfortunately need to include this quasi again...
*/
#include "RotTrans2D.h"


/*!*************************************
* Create a point given its two Cartesian coordinates.
* @param x The x coordinate.
* @param y The y coordinate
*/
Point2D::Point2D(double x, double y)
{
	coord[0] = x ;
	coord[1] = y ;
}

/*!*************************************
* Create a point at the origin of coordinates.
*/
Point2D::Point2D()
{
	coord[0] = coord[1]= 0. ;
}

/*!*************************************
* Create a point given its coordinate vector.
* @param xandy The x coordinate in [0], the y coordinate in [1].
*/
Point2D::Point2D(const double xandy[2])
{
#ifdef HAVE_MEMCPY
	memcpy(coord,xandy,sizeof(double[2])) ;
#else
	coord[0] = xandy[0] ;
	coord[1]= xandy[1] ;
#endif
}

/*!***********************************
* Compute the direction from this point to another.
* Equivalent description would be to constructe the vector to the other point minus this.
* @param oth The other point to connect this to.
*/
Point2D Point2D::to(const Point2D & oth) const
{
	/* construct the new vector by subtraction of the
	* cartesian coordinates.
	*/
	return Point2D(oth.coord[0]-coord[0], oth.coord[1]-coord[1]) ;
}

/*!***********************************
* @brief Add this point (as a vector) to another point.
* @param oth[in] The other point to attach this to.
* @return The point which has x and y coordinates which are sums of the x and
*   y coordinates of this and oth.
*/
Point2D Point2D::plus(const Point2D & oth) const
{
	/* construct the new vector by adding the
	* Cartesian coordinates.
	*/
	return Point2D(oth.coord[0]+coord[0], oth.coord[1]+coord[1]) ;
}

/*!***********************************
* Create a point that is rotated by 90 degrees ccw .
* @return A point with coordinates (-y,+x) if this =(+x,+y).
*/
Point2D Point2D::turn90() const
{
	return Point2D(-coord[1], coord[0]) ;
}

/*!***********************************
* Apply a rotation-translation to this point.
* The effect is the same as calling RotTrans2D::apply(const Point2D &) .
* @param[in] rt The rotation-translation operator to be applied.
* @return A point with coordinates rotated then translated by rt.
*/
Point2D Point2D::apply(const RotTrans2D & rt) const
{
	/* multply from the left by the matrix
	* (cos(alpha), - sin(alpha))
	* (sin(alpha), + cos(alpha))
	* and then translate by trans.
	*/
	const double x = rt.sc[1]* coord[0] - rt.sc[0]*coord[1] + rt.trans.coord[0] ;
	const double y = rt.sc[0]* coord[0] + rt.sc[1]*coord[1] + rt.trans.coord[1] ;
	return Point2D(x,y) ;

} /* apply */

/*!***********************************
* Compute the distance to another point.
* @param oth The other reference point.
* @return The distance between this and oth.
*  Square root of the sum of the squared differences in both Cartesian coordinates.
*/
double Point2D::dist(const Point2D & oth) const
{
	return hypot(oth.coord[0]-coord[0], oth.coord[1]-coord[1]) ;
}

/*!***********************************
* Compute the length of the vector from the origin to this point.
* @return The distance between this and the origin.
*/
double Point2D::len() const
{
	return hypot(coord[0], coord[1]) ;
}

/*!***********************************
* Compute the dot product of this point (as a vector) with another.
* @return The dot product this.x *oth.x +this.y*oth.y.
*/
double Point2D::dotprod(const Point2D &oth) const
{
	return coord[0]*oth.coord[0] + coord[1]*oth.coord[1] ;
}

#if 0
/*!****************************************
* Assignment operator.
* Superfluous since assignment on a member-by-member basis suffices.
*/
Point2D & Point2D::operator=( Point2D const & rhs)
{
	if ( this != & rhs)
	{
#ifdef HAVE_MEMCPY
		memcpy(coord,rhs.coord,sizeof(double[2])) ;
#else
		coord[0] = rhs.coord[0] ;
		coord[1] = rhs.coord[1] ;
#endif
	}
	return *this ;
}
#endif

/*!************************************
* @brief Subtract another point, interpreting both points as vectors.
*
* Conceptionally this is (up to a sign flip or up to a role reversal) the
* same as the function to(const Point2D &). So the same result would be
* obtained by the call rhs.to(*this).
* @param rhs The point to be subtracted.
* @return A point where coordinates are individually reduced by those of rhs.
*/
Point2D & Point2D::operator-= (const Point2D & rhs)
{
	coord[0] -= rhs.coord[0] ;
	coord[1] -= rhs.coord[1] ;
        return *this ;
} /* operator-= */

/*!************************************
* @brief Add another point, interpreting both points as vectors.
*
* @param rhs The point to be added.
* @return A point where coordinates are individually extended by those of rhs.
*/
Point2D & Point2D::operator+= (const Point2D & rhs)
{
	coord[0] += rhs.coord[0] ;
	coord[1] += rhs.coord[1] ;
        return *this ;
} /* operator+= */

/*!************************************
* @brief Add two points, interpreting both points as vectors.
*
* @param left The point to the left of the + sign.
* @param right The point to right of the + sign.
* @return A point with coordinates that are sums of coordinates of both arguments.
*/
Point2D operator+ (const Point2D & left, const Point2D & right)
{
	return Point2D(left.coord[0]+right.coord[0], left.coord[1]+right.coord[1]) ;
} /* operator+ */
