#include <cmath>
#include <iostream>

#include "Square2D.h"

/*!*************************************
* @brief Create a planar square given the first two adjacent vertices.
*
* The third and fourth vertex are defined by orthogonal
* extension off the baseline between the first and second vertex.
* @param pt1 The first vertex.
* @param pt2 The second vertex.
*/
Square2D::Square2D(const Point2D & pt1, const Point2D & pt2)
{
	vert[0] = pt1 ;
	vert[1] = pt2 ;
	/* the third vertex has the same distance to the 2nd as
	* the distance between the 1st and 2nd and is orthogonal
	* to the first edge.
	*/
	const Point2D edg0 = pt1.to(pt2) ;
	/*
	Point2D edg1(-edg0.coord[1],edg0.coord[0]) ;
	*/
	const Point2D edg1 = edg0.turn90() ;
	vert[2] = pt2.plus(edg1) ;
	vert[3] = pt1.plus(edg1) ;

	eLen = edg0.len() ;
} /* ctor */

/*!*************************************
* @brief Create a planar square given all four vertices.
*
* The four vertices are supposed to be given in ccs order.
* No checking on the length and mutual orthogonality of the edges is done!
* @param pt1 The first vertex.
* @param pt2 The second vertex.
* @param pt3 The third vertex.
* @param pt4 The fourth vertex.
*/
Square2D::Square2D(const Point2D & pt1, const Point2D & pt2, const Point2D & pt3, const Point2D & pt4)
{
	vert[0] = pt1 ;
	vert[1] = pt2 ;
	vert[2] = pt3 ;
	vert[3] = pt4 ;

	eLen = pt1.to(pt2).len() ;
} /* ctor */

/*!*************************************
* Create an empty square with four vertices all at the origin.
*/
Square2D::Square2D(): eLen(0.)
{
} /* default ctor */

/*!*******************************************
* @brief Generate the bounding box of this square.
* @return The quadrangular bounding box.
*/
Square2D::operator Bbox2D() const
{
	/* start with a bounding box from vertex 0 to vertex 2,
	* and extend this for the case that there is some nonzero tilt.
	*/
        Bbox2D bb= Bbox2D( vert[0], vert[2]) ;
	bb.include(vert[1]) ;
	bb.include(vert[3]) ;
	return bb ;
} /* operator::Bbox2D */


/*!**********************************
* Create a line that connects vertex number 'no' with the next vertex along the edge.
* @param no The edge number in the range 0 to 3.
* @return The line that connects point no with point no+1.
*  Edge 0 connects vertex 0 to 1.
*  Edge 1 connects vertex 1 to 2.
*  Edge 2 connects vertex 2 to 3.
*  Edge 3 connects vertex 3 to 0.
*/
Line2D Square2D::edge(int no) const
{
	return Line2D( vert[no %4], vert[(no+1) % 4]) ;
} /* edge */


/*!**********************************
* Compute the circumcircle.
* @return The circumcircle.
*  The perimeter of the circle touches all 4 vertices of the Square.
*/
Circle2D Square2D::circumC() const
{
	/* center position of the new circle. half way between vertex 0 and 2.
	*/
	double midxy[2] ;
	midxy[0] = (vert[0].coord[0]+vert[2].coord[0] )/2. ;
	midxy[1] = (vert[0].coord[1]+vert[2].coord[1] )/2. ;
	return Circle2D(midxy[0], midxy[1], eLen*M_SQRT2) ;
} /* circumC */

/*!*********************************
* Compute the area.
* @return The area. Square of the edge length
*/
double Square2D::area() const
{
	return eLen*eLen ;
} /* area */

/*!*********************************
* Dissect the square into two triangles along a diagonal.
* @return The two triangles.
*  The orientation of the triangles is individually the same
*  as the orientation of the square. We also guarantee that
*  the first two edges of each triangle are orthogonal, which implies
*  that the circumcircles of these are centered at the (common) longest edge.
*/
Tria2D * Square2D::dissect() const
{
	Tria2D * dis = new Tria2D[2] ;
//cout <<  __FILE__ << " " << __LINE__ << " dis0 " << endl ;
	dis[0] = Tria2D(vert[0],vert[1],vert[2]) ;
//cout <<  __FILE__ << " " << __LINE__ << " dis1 " << endl ;
	dis[1] = Tria2D(vert[2],vert[3],vert[0]) ;
//cout <<  __FILE__ << " " << __LINE__ << " dise " << endl ;
	return dis ;
} /* dissect */

/*!****************************************
* @brief Compute the square that is rotated-translated.
* @param[in] rt The rotation-translation to be applied
* @return The square defined by individual rotation of all 4 vertices of this.
*/
Square2D Square2D::apply(const RotTrans2D & rt) const
{
        /* Compute the coordinates of the first two of the old corners
        * under the tranformation.
        */
        Point2D imgCorn[2] ;
        imgCorn[0] = vert[0].apply(rt) ;
        imgCorn[1] = vert[1].apply(rt) ;

	/* todo: probably faster if 4 individual "apply" to the 4 vertices */
        return Square2D(imgCorn[0],imgCorn[1]) ;
} /* apply */

/*!**************************************************
* @brief Compute the common area (intersection) between this
* and another square.
* @return The vector of mutually non-intersecting pieces of the
*  common area of both polygons.
*  If there is no overlap, the vector that is returned is empty.
*/
vector<Tria2D> Square2D::intersect(const Square2D & oth) const
{
	/* Decompose oth into two triangles and forward the problem to
	* the sum of intersections of these two triangles with this.
	*/
	Tria2D * othtris = oth.dissect() ;

	/* The set of triangles to be returned; initially the intersections
	* of the first subtriangle.
	*/
	vector<Tria2D> intr = othtris[0].intersect(*this) ;
	/* And append the intersections of the second triangle
	* (which are not overlapping any of the triangles of the first set by construction).
	*/
	vector<Tria2D> intr2 = othtris[1].intersect(*this) ;
	intr.insert(intr.end(), intr2.begin(), intr2.end()) ;

	delete [] othtris ;
        return intr ;
} /* intersect */
