#include "config.h"

#include <iostream>

#ifdef HAVE_CMATH
#include <cmath>
#else
#include <math.h>
#endif

#include "Tria2D.h"
#include "Square2D.h"

/*!*************************************
* Create a planar triangle given its three vertices.
* @param pt1 The first vertex.
* @param pt2 The second vertex.
* @param pt3 The third vertex.
*/
Tria2D::Tria2D(const Point2D & pt1, const Point2D & pt2, const Point2D & pt3)
{
	vert[0] = pt1 ;
	vert[1] = pt2 ;
	vert[2] = pt3 ;

#if 0
	const Line2D l0(vert[0],vert[1]) ;
	bool ll  = l0.leftFrom( vert[2]) ;
	if ( ! ll)
	{
		cout << __FILE__ << " " << __LINE__ << " ctor " << vert[0].coord[0] << " " << vert[0].coord[1] << endl ;
		cout << __FILE__ << " " << __LINE__ << " ctor " << vert[1].coord[0] << " " << vert[1].coord[1] << endl ;
		cout << __FILE__ << " " << __LINE__ << " ctor " << vert[2].coord[0] << " " << vert[2].coord[1] << endl ;
	}
#endif
} /* ctor */

/*!*************************************
* Create an empty triangle with three vertices all at the origin.
*/
Tria2D::Tria2D()
{
} /* default ctor */

/*!**********************************
* Create a line that connects vertex number 'no' with the next vertex along the edge.
* @param no The edge number in the range 0 to 2.
* @return The line that connects point no with point no+1.
*  Edge 0 connects vertex 0 to 1.
*  Edge 1 connects vertex 1 to 2.
*  Edge 2 connects vertex 2 to 0.
*/
Line2D Tria2D::edge(int no) const
{
	return Line2D( vert[no %3], vert[(no+1) % 3]) ;
} /* edge */

/*!**********************************
* Compute the area of the triangle.
* @return The area of all points within the triangle.
*/
double Tria2D::area() const
{
	/* half of the absolute value of the cross product of any pair of two different edges.
	* By using e0 from 0 to 1 and e1 from 0 to 2 we ensure that rotating
	*/
	const Point2D e0 = vert[0].to(vert[1]) ;
	const Point2D e1 = vert[0].to(vert[2]) ;
	return 0.5*fabs(e0.coord[0]*e1.coord[1] - e1.coord[0]*e0.coord[1]) ;
} /* area */


/*!**********************************
* Compute the circumcircle.
* @return The circumcircle.
*  By definition, the perimeter of the circle touches all three
*  vertices of the triangle.
*/
Circle2D Tria2D::circumC() const
{
	const double disc = 2.0*(
		vert[0].coord[0]*(vert[1].coord[1]-vert[2].coord[2])
		+vert[1].coord[0]*(vert[2].coord[1]-vert[0].coord[2])
		+vert[2].coord[0]*(vert[0].coord[1]-vert[1].coord[2])
	) ;

	/* squared distance from the origin the vertex number 0 to 2
	*/
	double lensq[3] ;
	for(int pt= 0 ; pt < 3 ; pt++)
		lensq[pt] = vert[pt].coord[0]*vert[pt].coord[0]
			+vert[pt].coord[1]*vert[pt].coord[1] ;

	/* center position of the new circle
	*/
	double midxy[2] ;
	midxy[0] = lensq[0]*(vert[1].coord[1]-vert[2].coord[1])
		+ lensq[1]*(vert[2].coord[1]-vert[0].coord[1])
		+lensq[2]*(vert[0].coord[1]-vert[1].coord[1]) ;
	midxy[1] = lensq[0]*(vert[2].coord[0]-vert[1].coord[0])
		+ lensq[1]*(vert[0].coord[0]-vert[2].coord[0])
		+lensq[2]*(vert[0].coord[0]-vert[1].coord[0]) ;
	/* the radius is the distance between the center and any of the 
	* three vertices of the triangle.
	*/
	Point2D midCirc(midxy[0]/disc, midxy[1]/disc) ;
	return Circle2D(midCirc, midCirc.dist(vert[0]) ) ;
} /* circumC */

/*!**************************************************
* Compute the portion of this that is left from the infinite line lin.
* @param lin The line that cuts the 2D plane in half.
* @return The vector of mutually non-intersecting pieces of the
*  subarea of this that is to the left from the line.
*/
vector<Tria2D> Tria2D::chop(const Line2D & lin) const
{
	vector<Tria2D> intr;

	bool leri[3] ;
	/* Determine for each of the 3 vertices whether it is left from
	* the cutting edge (retained) or right from it (chopped off). This
	* gives 8 different cases of left-right associations.
	*/
	for (int p=0 ; p < 3; p++)
		leri[p] = lin.leftFrom( vert[p]) ;

	if ( leri[0] == false)
	{
		if ( leri[1] == false)
		{
			if ( leri[2] == true)
			{
				/* the infinitely extended line cuts through the edge from 
				* vertex 1 to vertex 2 and through the edge from vertex 2
				* to vertex 0. The result is a single smaller triangle.
				*/
				Point2D intrs0 = lin.intersectInf(edge(2)) ;
				Point2D intrs1 = lin.intersectInf(edge(1)) ;
				Tria2D subtria(intrs0, intrs1,vert[2]) ;
				intr.push_back(subtria) ;
			}
			/* otherwise all to the right, and so is the entire oth,
			* which means nothing to do and the return vector is empty.
			*/
		}
		else
		{
			if ( leri[2] == true)
			{
				/* the infinitely extended line cuts through the edge from 
				* vertex 0 to vertex 1 and through the edge from vertex 2
				* to vertex 0. The result is a quadrangle which is returned as 2 triangles
				* by dissection along one (arbitrary of 2) diagonal.
				*/
				Point2D intrs0 = lin.intersectInf(edge(2)) ;
				Point2D intrs1 = lin.intersectInf(edge(0)) ;
				Tria2D subtria1(intrs0, intrs1,vert[1]) ;
				intr.push_back(subtria1) ;
				Tria2D subtria2(intrs0, vert[1], vert[2]) ;
				intr.push_back(subtria2) ;
			}
			else
			{
				/* vertex 1 is the only vertex to the left, so
				* the line cuts through the edge from vertes 0 to vertex 1
				* and through the edge from vertex 1 to vertex 2.
				* The result is a single triangle with vert[1] as one vertex.
				*/
				Point2D intrs0 = lin.intersectInf(edge(1)) ;
				Point2D intrs1 = lin.intersectInf(edge(0)) ;
				Tria2D subtria(intrs0, intrs1,vert[1]) ;
				intr.push_back(subtria) ;
			}
		}
	}
	else
	{
		if ( leri[1] == false)
		{
			if ( leri[2] == true)
			{
				/* vert[0] is to the left, vert[1] to the right and vert[2]
				* to the left, so as above th eline cute through the edge
				* from vert[1] to vert[2] and through the edge from vert[0]
				* to vert[1]. The quadrangle to the left is split into
				* two triangles.
				*/
				Point2D intrs0 = lin.intersectInf(edge(1)) ;
				Point2D intrs1 = lin.intersectInf(edge(0)) ;
				Tria2D subtria1(intrs1, intrs0,vert[2]) ;
				intr.push_back(subtria1) ;
				Tria2D subtria2(intrs1, vert[2], vert[0]) ;
				intr.push_back(subtria2) ;
			}
			else
			{
				/* vert[0] is to the left, vert[1] to the right and vert[2]
				* to the right. The line cuts through the edge from
				* vert[2] to vert[0] and through the edge from vert[0]
				* to vert[1]. A single triangle with vert[0] as an edge is returned.
				*/
				Point2D intrs0 = lin.intersectInf(edge(2)) ;
				Point2D intrs1 = lin.intersectInf(edge(0)) ;
				Tria2D subtria(intrs1, intrs0,vert[0]) ;
				intr.push_back(subtria) ;
			}
		}
		else
		{
			if ( leri[2] == true)
			{
				/* all three vertices are to the left.
				* The full triangle is returned.
				*/
				intr.push_back(*this) ;
			}
			else
			{
				/* vert[0] is to the left, vert[1] to the left and vert[2]
				* to the right. The line cuts through the edge from
				* vert[2] to vert[0] and through the edge from vert[1]
				* to vert[2]. A quadrangle split into two triangles is returned.
				*/
				Point2D intrs0 = lin.intersectInf(edge(2)) ;
				Point2D intrs1 = lin.intersectInf(edge(1)) ;
				Tria2D subtria1(intrs1, intrs0,vert[0]) ;
				intr.push_back(subtria1) ;
				Tria2D subtria2(intrs1, vert[0], vert[1]) ;
				intr.push_back(subtria2) ;
			}
		}
	}
	return intr ;
} /* chop */

/*!**************************************************
* @brief Compute the common portion (intersection) between this
* and another triangle.
* @return The vector of mutually non-intersecting pieces of the
*  common area of both triangles.
*  If they do not overlap, the vector is empty.
*/
vector<Tria2D> Tria2D::intersect(const Tria2D & oth) const
{
	/* The set of triangles to be returned; initially empty
	*/
	vector<Tria2D> intr;

#if 0
	/* quick look: if the circumscribing circles of this and oth don't intersect,
	* the triangles don't intersect either, and intr remains empty.
	*/
	Circle2D thisC = circumC() ;
	Circle2D othC = oth.circumC() ;
	if ( thisC.overlap(othC) )
#endif
	{
		/* Use the three edges of oth as scissors that cut through the edges
		* of this. Start with the first edge and then use the other two.
		*/
		Line2D cutedg = oth.edge(0) ;
		vector<Tria2D> left0 = chop(cutedg) ;

		cutedg = oth.edge(1) ;
		vector<Tria2D> left1 ;
		for( int t=0 ; t < left0.size() ; t++)
		{
			vector<Tria2D> tmp = left0[t].chop(cutedg) ;
			left1.insert(left1.end(), tmp.begin(), tmp.end()) ;
		}

		cutedg = oth.edge(2) ;
		for( int t=0 ; t < left1.size() ; t++)
		{
			vector<Tria2D> tmp = left1[t].chop(cutedg) ;
			intr.insert(intr.end(), tmp.begin(), tmp.end()) ;
		}
	}
	return intr ;
} /* intersect */

/*!**************************************************
* @brief Compute the common area (intersection) between this
* and a square.
* @return The vector of mutually non-intersecting pieces of the
*  common area of both polygons..
*  If they do not overlap, the vector is empty.
*/
vector<Tria2D> Tria2D::intersect(const Square2D & oth) const
{
	/* The set of triangles to be returned; initially empty
	*/
	vector<Tria2D> intr;

#if 0
	/* quick look: if the circumscribing circles of this and oth don't intersect,
	* the triangles don't intersect either, and intr remains empty.
	* In the application here, this actually slows down the program because
	* the pre-selection of the few target squares around the rotated source square
	* means that this criterion is rarely valid, such that computing this
	* check is not beneficial.
	*/
	Circle2D thisC = circumC() ;
	Circle2D othC = oth.circumC() ;
	if ( thisC.overlap(othC) )
#endif
	{
		/* Use the four edges of the quadrangle a scissors that cut through the edges
		* of this. Start with first edge.
		*/
		Line2D cutedg = oth.edge(0) ;
		vector<Tria2D> left0 = chop(cutedg) ;
		if ( left0.size() ==0 )
			return intr ;

		cutedg = oth.edge(1) ;
		vector<Tria2D> left1 ;
		for( int t=0 ; t < left0.size() ; t++)
		{
			vector<Tria2D> tmp = left0[t].chop(cutedg) ;
			left1.insert(left1.end(), tmp.begin(), tmp.end()) ;
		}
		if ( left1.size() ==0 )
			return intr ;

		cutedg = oth.edge(2) ;
		vector<Tria2D> left2 ;
		for( int t=0 ; t < left1.size() ; t++)
		{
			vector<Tria2D> tmp = left1[t].chop(cutedg) ;
			left2.insert(left2.end(), tmp.begin(), tmp.end()) ;
		}
		if ( left2.size() ==0 )
			return intr ;

		cutedg = oth.edge(3) ;
		for( int t=0 ; t < left2.size() ; t++)
		{
			vector<Tria2D> tmp = left2[t].chop(cutedg) ;
			intr.insert(intr.end(), tmp.begin(), tmp.end()) ;
		}
	}
	return intr ;
} /* intersect */
