/*
* MPIA - GEIRS project
*
* "@(#) $Id$"
*
*/

#include "config.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <cstdlib>
#include <string>
#include <cstring>
#include <vector>

#include "Window.h"

using namespace std ;
using namespace CCfits ;

/** 
* @param[in] fitsname The name of the FITS file to be read.
* @param[in] gp The width of the gap between individual chips in the mosaic in units of pixels.
* @param[in] csiz The dimension along each edge of each chip in pixels.
* @since 2012-10-23
* @author Richard J. Mathar
*/
Window::Window(string fitsIname, int gp, int csiz) : iname(fitsIname), gap(gp), chpsz(csiz) , pixsc(0.)
{
	/* open the input fits file (for GEIRS, this is one of the ????*_win.fits
	* or the single ????.fits file).
	*/
	ifits = new FITS(iname, CCfits::Read) ;
#ifdef DEBUG
	FITS::setVerboseMode(true) ;
#endif
	/* gather the information on subwindow location and size, mosaic size
	* but do not read the ADU's yet.
	*/
	init(true) ;

	/* read the pixels
	*/
	readPix(true) ;
} /* Window::Window */

/** 
* @param[in] fitsname The name of the FITS file to be read.
* @param[in] gp The width of the gap between individual chips in the mosaic in units of pixels.
* @param[in] csiz The dimension along each edge of each chip in pixels.
* @since 2012-10-23
* @author Richard J. Mathar
*/
Window::Window(char *fitsIname, int gp, int csiz) : iname(fitsIname), gap(gp), chpsz(csiz) , pixsc(0.)
{
	/* open the input fits file (for GEIRS, this is one of the ????*_win.fits
	* or the single ????.fits file).
	*/
	ifits = new FITS(iname, CCfits::Read) ;
#ifdef DEBUG
	FITS::setVerboseMode(true) ;
#endif
	/* gather the information on subwindow location and size, mosaic size
	* but do not read the ADU's yet.
	*/
	init(false) ;

	/* read the pixels
	*/
	readPix(false) ;
} /* Window::Window */

/** Destructor.
*/
Window::~Window()
{
	/* remove the allocated resoures with the ctor of ifits within the ctor
	*/
	delete ifits ;
	ifits =0 ;
} /* Window::dtor */

/** Delete the file that contains the data used on construction.
* Obviously, this is a risky operation and may lead to loss of data.
* @param[in] verbose If true, print names of files deleted to stdout.
* @since 2012-11-20
* @author Richard J. Mathar
*/
void Window::rmFiles(bool verbose)
{
	unlink(iname.c_str()) ;
	if (verbose)
		cout << "... " << iname.c_str() << " deleted" << endl ;
} /* rmFiles */

/** Scan the bare primary information in the FITS header.
* Read INSTRUME to have some input on the mosaic size.
* Read DETSIZE which is the preferred way to have some input on the mosaic size.
* Read SAVEAREA (GEIRS specific, deprecated) and TRIMSIZE to determine the
* window location
* @author Richard J. Mathar
*/
void Window::init(bool verbose)
{
	PHDU & iex = ifits->pHDU() ;

	iex.readAllKeys() ;

	naxes = iex.axes() ;

	map<String, Keyword*> & allkw = iex.keyWord() ;

	detsize[0] = detsize[2] = 1 ;
	detsize[1] = detsize[3] = 2048 ;
	/* copy most header keyword lines...
	*/
	map<String, Keyword*>::iterator kitr = allkw.begin() ;
	for ( ; kitr != allkw.end() ; kitr++)
	{
		const string & knam = (*kitr).first;
		const Keyword * thisk = (*kitr).second ;
		if ( knam == "PIXSCALE" || knam == "PIXSCAL" )
		{
			/* remember the pixel scale so we can build the CD matrix later
			*/
			thisk->value(pixsc) ;
			if (verbose)
				cout << "PIXSCALE in " << iname << " " << pixsc << " arcsec/px " << endl ;
			pixsc /= 3600.0 ;
		}
		else if ( knam == "INSTRUME" )
		{
			string instr ;
			/* backup in case that DETSIZE is not provided
			*/
			thisk->value(instr) ;

			/* Panic is 2x2 mosaic of 2k chips, Carmenes is
			* 2x1 mosaic of 2k chips. Linc-Nirvana is one 2k chip,
			* but less defined as a INSTRUME and covered by
			* the default setting above the loop.
			* Laica is a 2x2 mosaic of 4k chips.
			*/
			if ( instr.find_first_of("Panic") != string::npos || instr.find_first_of("PANIC") != string::npos)
				detsize[1] = detsize[3] = 4096 ;
			else if ( instr.find_first_of("Carmen") != string::npos || instr.find_first_of("CARMEN") != string::npos)
			{
				detsize[1] = 4096 ;
				detsize[3] = 2048 ;
			}
			else if ( instr.find_first_of("Laica") != string::npos || instr.find_first_of("LAICA") != string::npos)
			{
				detsize[1] = detsize[3] = 8192 ;
			}
		}
		else if ( knam == "SAVEAREA" || knam == "TRIMSIZE" || knam == "DETSEC")
		{
			/* of the form [llx:lly,wx:wy]
			*/
			string tsiz ;
			thisk->value(tsiz) ;
			fourPix(tsiz,trimsizeFus,false) ;
			if (verbose)
				cout << knam << " in " << iname << " " << trimsizeFus[0] << " " << trimsizeFus[1] << " "
				 << trimsizeFus[2] << " " << trimsizeFus[3] << endl ;
		}
		else if ( knam == "DETSIZE" )
		{
			/* Of the form [llx:lly,wx:wy]
			* It is a deficiency of the current GEIRS version that
			* DETSIZE is not put into the FITS header.
			*/
			string tsiz ;
			thisk->value(tsiz) ;
			fourPix(tsiz,detsize,false) ;
			if (verbose)
				cout << knam << " in " << iname << " " << detsize[0] << " " << detsize[1] << " "
				 << detsize[2] << " " << detsize[3] << endl ;
		}
	}

	/* up to here, the trimsizeFus[] and detsize[] coordinates were 1-based,
	* as in FITS and FORTRAN. Move on to base-0 C/C++ convention by subtracting 1 from the lower x and y coods.
	* The upper cordinate then is exclusive.
	*/
	trimsizeFus[0] -= 1 ;
	trimsizeFus[2] -= 1 ;
	detsize[0] -= 1 ;
	detsize[2] -= 1 ;

	/* native location on the rotated detector image not yet known
	*/
	trimsize[0] = trimsize[1] = trimsize[2] = trimsize[3]= 0 ;
} /* Window::init */

/** Read the pixel data of the primary data unit (image).
* @param[in] verbose If true, write some output to stdout.
* @since 2012-10-18
* @author Richard J. Mathar
*/
void Window::readPix(bool verbose)
{
	PHDU & iex = ifits->pHDU() ;
	bits = iex.bitpix() ;

	/* the x and y bits length of the original GEIRS data, third dimension >=1 if cubes.
	*/
	long origdsec[3] = {0L,0L,1L} ;
	for(int i=0 ; i < naxes ; i++)
		origdsec[i] = iex.axis(i) ;

	if (verbose)
	{
		cout << "Original " << iname << " " << origdsec[0] << " by " << origdsec[1] <<  " by " << origdsec[2] << " pixels." << endl ;
		// cout << "\t0-based DATASEC " << trimsizeFus[0] << " " << trimsizeFus[1] << " " << trimsizeFus[2]<< " " << trimsizeFus[3] << endl ;
	}

	/* already done with init()
	* iex.readAllKeys() ;
	*/

	/* not documented with CCfits (which does not do any scaling) but with cfitsio
	* is that BSCALE and BZERO have been applied by the read (and will be applied by
	* the write) functions on the fly.
	*/
	switch(bits)
	{
	case USHORT_IMG:
		iex.read(usarr,1,origdsec[0]*origdsec[1]*origdsec[2]) ;
		break;
	case SHORT_IMG:
		iex.read(sarr,1,origdsec[0]*origdsec[1]*origdsec[2]) ;
		break;
	case LONG_IMG:
		iex.read(iarr,1,origdsec[0]*origdsec[1]*origdsec[2]) ;
		break;
	case FLOAT_IMG:
		iex.read(farr,1,origdsec[0]*origdsec[1]*origdsec[2]) ;
		break;
	case BYTE_IMG:
	case LONGLONG_IMG:
	case DOUBLE_IMG:
		break;
	}

} /* Window::readPix */

/** Generate a <INSTRU>.YYYY-MM-DDTHH-MM-SS.sss.fits string name.
* @since 2012-11-20
* @author Richard J. Mathar
*/
string Window::canonName()
{
	PHDU & iex = ifits->pHDU() ;

	map<String, Keyword*> & allkw = iex.keyWord() ;

	/* the YYYY-DD... string of either DATE-OBS (preferred) or DATE (backup)
	*/
	string dtstring ;

	/* The INSTRU string in the header
	*/
	string instr ;

	map<String, Keyword*>::iterator kitr = allkw.begin() ;
	for ( ; kitr != allkw.end() ; kitr++)
	{
		const string & knam = (*kitr).first;
		const Keyword * thisk = (*kitr).second ;
		if ( knam == "DATE-OBS" || ( knam == "DATE" && dtstring.length() == 0) )
		{
			string tmp ;
			thisk->value(tmp) ;
			dtstring = trimws(tmp) ;
		}
		else if ( knam == "INSTRUME" )
		{
			string tmp ;
			thisk->value(tmp) ;
			instr = trimws(tmp) ;
		}
	}
	return instr + "." + dtstring + ".fits" ;

} /* Window::canonName */

#define GEIRS2PANIC_WINDOW_TRIM 4

/** Decide whether a pixel is one of the reference pixel in one of the 4 frame borders.
* This supports elimination of the reference pixels on Hawaii2 detectors.
* @param[in] xin The x coordinate in the range 0 to xchipsize-1 (inclusive)
*   Independent of windowing, this coordinates refers to x=0 as the lower left edge
*   coordinate of the entire mosaic.
* @param[in] yin The y coordinate in the range 0 to ychipsize-1 (inclusive)
*   Independent of windowing, this coordinates refers to y=0 as the lower left
*   edge coordinate the entire mosaic.
* @param[in] xchipsize The number of pixels along x.
* @param[in] ychipsize The number of pixels along y.
* @return true if (xin,yin) lie in one of the 4-pixel white frames around the full frame.
*/
bool Window::isRef(int xin, int yin, int xchipsize, int ychipsize)
{
	/* reduce values modulo the chipsize to a cross in the range near zero
	*/
	while ( xin > xchipsize/2)
		xin -= xchipsize ;
	while ( yin > ychipsize/2)
		yin -= ychipsize ;
	/* is a reference pixel if in the range -4 to +3 in either coordinate
	*/
	return ( (xin >= -GEIRS2PANIC_WINDOW_TRIM && xin < GEIRS2PANIC_WINDOW_TRIM ) 
		|| (yin >= -GEIRS2PANIC_WINDOW_TRIM && yin < GEIRS2PANIC_WINDOW_TRIM ) ) ;
} /* isRef */

/** Decide whether a pixel is one of the reference pixel in one of the 4 frame borders.
* This supports elimination of the reference pixels on Hawaii2 detectors.
* @param[in] xin The x coordinate in the range 0 to xchipsize-1 (inclusive)
*   Independent of windowing, this coordinates refers to x=0 as the lower left edge
*   coordinate of the entire mosaic.
* @param[in] yin The y coordinate in the range 0 to ychipsize-1 (inclusive)
*   Independent of windowing, this coordinates refers to y=0 as the lower left
*   edge coordinate the entire mosaic.
* @return true if (xin,yin) lie in one of the 4-pixel white frames around the full frame.
*/
bool Window::isRef(int xin, int yin) const
{
	/* reduce values modulo the chipsize to a cross in the range near zero
	*/
	while ( xin > chpsz/2)
		xin -= chpsz ;
	while ( yin > chpsz/2)
		yin -= chpsz ;
	/* is a reference pixel if in the range -4 to +3 in either coordinate
	*/
	return ( (xin >= -GEIRS2PANIC_WINDOW_TRIM && xin < GEIRS2PANIC_WINDOW_TRIM ) 
		|| (yin >= -GEIRS2PANIC_WINDOW_TRIM && yin < GEIRS2PANIC_WINDOW_TRIM ) ) ;
} /* isRef */

/** Fill any pixel in one of the four pixel borders (within the 4pixel limit) by the blank.
* This supports elimination of the reference pixels on Hawaii2 detectors.
* @param[in] blank The value to replace all reference values.
* @since 2012-11-19
*/
void Window::blankRef(int blank)
{
	/* loop over all detector double edges horizontally
	*/
	const int winsx = sizeX() ;
	const int winsy = sizeY() ;
	double coo[2] ;
	for(int box =0 ; box <= detsize[1] ; box += chpsz)
	{
		/* 8 two-ways pixel lanes
		*/
		for(int delx = -GEIRS2PANIC_WINDOW_TRIM ; delx < GEIRS2PANIC_WINDOW_TRIM ; delx++)
		{
			/* handle only if this vertical stripe hits the subwindow anywhere
			*/
			if ( box+delx >= trimsizeFus[0] && box+delx < trimsizeFus[1])
			{
				coo[0] = box+delx-trimsizeFus[0] ;
				for( coo[1] = 0 ; coo[1] < winsy ; coo[1]++)
				{
					/* null that pixel */
					if ( farr.size() > 0)
						farr[coo[0]+winsx*coo[1]] = blank ;
					else if ( iarr.size() > 0 )
						iarr[coo[0]+winsx*coo[1]] = blank ;
					else if ( usarr.size() > 0 )
						usarr[coo[0]+winsx*coo[1]] = blank ;
				}
			}
		}
	}
	for(int boy =0 ; boy <= detsize[3] ; boy += chpsz)
	{
		/* 8 two-ways pixel lanes
		*/
		for(int dely = -GEIRS2PANIC_WINDOW_TRIM ; dely < GEIRS2PANIC_WINDOW_TRIM ; dely++)
		{
			/* handle only if this vertical stripe hits the subwindow anywhere
			*/
			if ( boy+dely >= trimsizeFus[2] && boy+dely < trimsizeFus[3])
			{
				coo[1] = boy+dely-trimsizeFus[2] ;
				for( coo[0] = 0 ; coo[0] < winsx ; coo[0]++)
				{
					/* null that pixel */
					if ( farr.size() > 0)
						farr[coo[0]+winsx*coo[1]] = blank ;
					else if ( iarr.size() > 0 )
						iarr[coo[0]+winsx*coo[1]] = blank ;
					else if ( usarr.size() > 0 )
						usarr[coo[0]+winsx*coo[1]] = blank ;
				}
			}
		}
	}
} /* blankRef */

/** Extract the [xl:xh,yl:yh] four parameters from the bracket.
* @param[in] tsiz with four integer values, colon and comma-separated.
*   The interpretation for relwidths=true is that the value of ranges in x is from xl (inclusive) to xl+xh (exclusive)
*   and in y from yl (inclusive) to yl+yh (exclusive).
*   The interpretation for relwidths=false is that the value of ranges in x is from xl (inclusive) to xh (exclusive)
*   and in y from yl (inclusive) to yh (exclusive).
* @param[out] coo The four coordinates xl, xh, yl, yh. If the format of the
*   string does not provide the two brackets, two colons and comma, the values are
*   not changed.
* @param[in] relwidths If true, the 2nd pair of coordinates is interpreted
*   as relative sizes (widths). If false, the pair is in the same absolute
*   coordinate system as the first pair.
* @since 2012-11-06
*/
void Window::fourPix(const string & tsiz, int coo[4],bool relwidths)
{
	/* find first bracket, last bracket, first and last colon, and comma */
	size_t fbrak = tsiz.find_first_of("[") ;
	size_t lbrak = tsiz.find_last_of("]") ;
	size_t fcol = tsiz.find_first_of(":") ;
	size_t lcol = tsiz.find_last_of(":") ;
	size_t com = tsiz.find_first_of(",") ;

	if ( fbrak != string::npos && lbrak != string::npos && fcol != string::npos
		&& lcol != string::npos && com != string::npos )
	{
		/* first number between fbrak and fcol
		*/
		sscanf(tsiz.substr(fbrak+1,fcol-fbrak-1).c_str(),"%d",&coo[0]) ;

		/* second number between fcol and com
		*/
		sscanf(tsiz.substr(fcol+1,com-fcol-1).c_str(),"%d",&coo[1]) ;

		/* third number between com and lcol
		*/
		sscanf(tsiz.substr(com+1,lcol-com-1).c_str(),"%d",&coo[2]) ;

		/* fourth number between lcol and lbrak
		*/
		sscanf(tsiz.substr(lcol+1,lbrak-lcol-1).c_str(),"%d",&coo[3]) ;

		if ( relwidths)
		{
			/* the original last parameters are interpreted as widths, so
			* we convert them to upper right integer absolute coordinates,
			* assuming this is the more common use.
			*/
			coo[1] += coo[0]-1 ;
			coo[3] += coo[2]-1 ;
		}
	}
} /* fourPix */

/** Convert the coordinate in the fused (zero-gap) windowed coordinate system to non-windowed.
* @param[in,out] coo On input, the x and y coordinate in the window and fused coordinate system.
*  On output, the x and y coordinate in the window and fused coordinate system.
*/
inline void Window::cooGeirsNoWin(int coo[2])
{
	coo[0] += trimsizeFus[0] ;
	coo[1] += trimsizeFus[2] ;
} /* cooGeirsNoWin */

/** Convert the coordinate in non-windowed coordinate system to windowed.
* @param[in,out] coo On input, the x and y coordinate in the windowed spliced coordinate system.
*  On output, the x and y coordinate in the global spliced coordinate system.
*/
inline void Window::cooToWin(int coo[2])
{
	coo[0] -= trimsize[0] ;
	coo[1] -= trimsize[2] ;
} /* cooToWin */

/** Compute the chip number in the fused (zero-gap) global mosaic coordinate system.
* @param[in] coo 
* @param[out] idx
* @since 2012-11-07
*/
inline void Window::detIdx(const int coo[2], int idx[2])
{
	idx[0] = coo[0]/chpsz ;
	idx[1] = coo[1]/chpsz ;
} /* detIdx */

/** Map a pixel in the input rectangular array to a pixel in the output array.
* @param[in,out] coo [0] the x and [1] the y coordinate before and after the rotation.
*  On output, the two coordinates may have picked up any of the 4 sign combinations.
*  So this is a pure rotation on the Gaussian integer coordinate system.
*  No shift is involved to translate the coordinates back to some sort of first quadrant.
* @param[in] rotRight90 Count of right rotations in units of 90 degrees.
*  This value is used modulo 4, so effectively between 1 and 3. 0 has no effect,
*  unless combined with the flip.
* @param[in] flip Request a flip around the y-axis. This equals sign toggling of x-coordinates
*  after any rotation.
*/
void Window::pixRot(int coo[2], int rotRight90, bool flip)
{
	int tmp ;
	switch(rotRight90 % 4 )
	{
	case 1 :
		/* 90 degrees to the right, 270 degrees to the left */
		tmp = coo[1] ;
		coo[1] = -coo[0] ;
		coo[0] = tmp ;
		break;
	case 2 :
		coo[0] = -coo[0] ;
		coo[1] = -coo[1] ;
		break;
	case 3 :
		/* 270 degrees ot the right, 90 degrees to the left
		*/
		tmp = coo[1] ;
		coo[1] = coo[0] ;
		coo[0] = -tmp ;
		break;
		/* noop for case 0*/
	}

	if ( flip )
		coo[0] = -coo[0] ;
} /* pixRot */

/** Convert the coordinate from fused (zero-gap) windowed to global mosaic coordinate (spliced) system.
* @param[in,out] coo 
* @since 2012-11-07
*/
void Window::cooGeirsWinToSplic(int coo[2])
{
	/* translate from fused windowed to fused non-windowed mosaic coordinate system
	*/
	cooGeirsNoWin(coo) ;

	/* get chip number, which is two coordinates in x and y both starting
	* at 0 if this is the lower left chip.
	*/
	int idx[2] ;
	detIdx(coo, idx) ;

	/* Move on to the global, spliced mosaic cordinates.
	* Add one gap pixel count for each increase by one
	*/
	coo[0] += gap*idx[0] ;
	coo[1] += gap*idx[1] ;

} /* cooGeirsWinToSpli */

/** Convert the coordinate from fused (zero-gap) windowed to global mosaic coordinate (spliced) system.
* @param[in,out] coo 
* @since 2012-11-07
*/
void Window::cooGeirsWinToSplicRot(int coo[2], const int rotRight90, const bool flip)
{
	/* convert windowed fused coordinates by translation/chopping to global non-windowed sliced
	* coordinates
	*/
	cooGeirsWinToSplic(coo) ;
	/* rotate within the global non-windowed sliced coordinates
	*/
	pixRot(coo, rotRight90, flip ) ;
} /* cooGeirsWinToSpli */

/** Convert the coordinate from fused (zero-gap) windowed to mosaic windowed/rotated coordinate (spliced) system.
* @param[in,out] coo 
* @since 2012-11-07
*/
void Window::cooGeirsWinToSplicRotWin(int coo[2], const int rotRight90, const bool flip)
{
	/* determine coordinates in the fused global mosaic coordinate system.
	*/
	cooGeirsWinToSplicRot(coo, rotRight90, flip) ;

	/* and subtract/relate to the window in that global coordinate system.
	*/
	cooToWin(coo) ;
} /* cooGeirsWinToSpliRotWin */

/** Compute the window location after splicing and optionial flip-rotation.
* @param[in] rotRight90 Number of right-90 rotations requested.
* @param[in] flip If true, perform antoher right-left flip around the y-coordinate.
*/
void Window::win2win(const int rotRight90, const bool flip)
{
	/* take the 4 window corners (the interior ones) in the fused windowed coordinate system.
	* The interior ones ensure that in some marginal cases, where these are adjacent to a gap,
	* are not implicitly encompassing the adjacent gap on output.
	*/
	int coo[4][2] ;
	coo[0][0] = 0 ; /* lower left x in fused windowed system */
	coo[0][1] = 0 ; /* lower left y in fused windowed system */
	coo[1][0] = trimsizeFus[1]-1-trimsizeFus[0] ; /* lower right x in fused windowed system */
	coo[1][1] = 0 ; /* lower right y in fused windowed system */
	coo[2][0] = coo[1][0] ; /* upper right x in fused windowed system */
	coo[2][1] = trimsizeFus[3]-1-trimsizeFus[2] ; /* upper right y in fused windowed system */
	coo[3][0] = 0 ; /* upper left x in fused windowed system */
	coo[3][1] = coo[2][1] ; /* upper left y in fused windowed system */

	/* for each of these four points, expand to the global spliced
	* coordinate system, and rotate/flip if requested.
	*/
	for (int p=0; p < 4 ; p++)
		cooGeirsWinToSplicRot(coo[p],rotRight90,flip) ;

	/* new lower x coordinate is the minimum of the x
	*/
	trimsize[0] = min(min( coo[0][0], coo[1][0]), min(coo[2][0],coo[3][0])) ;

	/* new lower y coordinate is the minimum of the y
	*/
	trimsize[2] = min(min( coo[0][1], coo[1][1]), min(coo[2][1],coo[3][1])) ;

	/* new right x coordinate is the maximum of the x, plus 1 to use exclusion
	*/
	trimsize[1] = 1+max(max( coo[0][0], coo[1][0]), max(coo[2][0],coo[3][0])) ;

	/* new upper y coordinate is the maximum of the y, plus 1 to use exclusion
	*/
	trimsize[3] = 1+max(max( coo[0][1], coo[1][1]), max(coo[2][1],coo[3][1])) ;
} /* win2win */

/** The width along x in pixel units.
* @return The horizontal size of the window in pixels. This refers to the
*  original orientation, not the optionally rotated or flipped one.
* @since 2012-11-07
*/
int Window::sizeX() const
{
	return trimsizeFus[1]-trimsizeFus[0]  ;
} /* sizeX */

/** The width along y in pixel units.
* @return The vertical size of the window in pixels. This refers to the
*  original orientation, not the optionally rotated or flipped one.
* @since 2012-11-07
*/
int Window::sizeY() const
{
	return trimsizeFus[3]-trimsizeFus[2]  ;
} /* sizeY */

/** Strip leading and trailing white space from a string.
* @param[in] instring The initial string from which nonprintable leading and
*   trailing characters or white space (tabs, blanks) ought to be removed.
* @return The string with leading and trailing blanks, tabs etc removed.
*/
string Window::trimws(const string & instring)
{
        /* this is mad actually more strict, using various iswhite()
        string::size_type head = instring.find_first_not_of(" \t\n\r") ;
        */
        string::size_type head =0 ;
        for( ; head < instring.length() ; head++)
        {
                /* this is liberal and allows punctuation (dots, commas..) in the names
                if ( isalpha(instring[head]) || isdigit(instring[head]) || ispunct(instring[head]) )
                */
                if ( isalnum(instring[head]) || ispunct(instring[head]) )
                        break ;
        }

        /* return empty string if everything is white space or non-printable
        */
        if ( head == string::npos)
                return string("") ;
        /*
        string::size_type tail = instring.find_last_not_of(" \t\n\r") ;
        */
        string::size_type tail = instring.length()-1 ;
        for( ; tail >= 0 ; tail--)
        {
                /* this is liberal and allows punctuation (dots, commas..) in the names
                */
                if ( isalnum(instring[tail]) || ispunct(instring[tail]) )
                        break ;
        }
        return instring.substr(head,tail+1) ;
} /* Window::trimws */

