#ifndef WINDOW_H
#define WINDOW_H
/*
* MPIA - GEIRS project
*
* "@(#) $Id$"
*
*/

#include <string>

/** The CCfits library contained in
* <a href="http://heasarc.gsfc.nasa.gov/docs/software/fitsio/ccfits/"> CCfits 2.4</a>
*/
#include <CCfits/CCfits>

using namespace std ;
using namespace CCfits ;

/** 
* @since 2012-11-08
* @author Richard J. Mathar
*/
class Window {

public:
	/** Name of the input file.
	* The principal use is to look into the primary header for the SAVEARE (TRIMSIZE) keyword
	* that relocates this window in the coordinate system of the mosaic of fused chips.
	*/
	string iname ;

	/** Gap (measured in pixels) between the chips of the mosaic.
	* This is considered the same horizontally and vertically.
	*/
	int gap ;

	/** Size (measured in pixels) of the single chip of the mosaic.
	* This is only considering quadratic chips (same x and y extent)
	*/
	int chpsz ; 

	/** Size (measured in pixels) of the fused array of chips.
	* This is the product of chip count and chpsz in the horizontal and vertical directions.
	* The interpretation is that the xrange is from [0] to [1] (exclusive), the y range
	* from [2] to [3] (exclusive) and [0]=[2]=0 with the usual 0-based C/C++/Java convention.
	*/
	int detsize[4] ; 

	/** Window origin and extension in the coordinate system of the fused (gapless) mosaic.
	* The four components of the array are following the same convention as detsize.
	*/
	int trimsizeFus[4] ;

	/** Window origin and extension in the coordinate system of the spliced mosaic.
	* The four components of the array are following the same convention as detsize.
	* If some subareas of the window are located in any but the lower-left chip, the components
	* will differ (and be larger than) the associated components in trimsizeFus.
	*/
	int trimsize[4] ;

	/** The FITS file that will be read/scanned.
	*/
	FITS * ifits ;

	/** array of pixels in floating point units. Empty if this is not a floating point type.
	*/
        valarray<float> farr  ;

	/** Array of pixels in 32bit integer units. Empty if this is not a integer type.
	*/
        valarray<int> iarr ;

	/** Array of pixels in 16bit integer units. Empty if this is not a unsigned short integer type.
	*/
        valarray<unsigned short> usarr ;

	/** Array of pixels in 16bit integer units. Empty if this is not a short integer type.
	*/
        valarray<unsigned short> sarr ;

#if 0
	/** Minimum and maximum of the data. Not used because it is not initially obvious
	* whether the refrence pixels on the edges are included.
	*/
	double mima[2] ;
#endif

	/** Pixel scale in units of degrees on the sky.
	* To be used to convert pixels to WCS coordinates.
	*/
	float pixsc ;

	/** The indication of image type (integer, short, float...)
	*/
	long bits ;

	/** The indication of dimension of the image (2=flat, 3=cubes,...)
	*/
	long naxes ;

	Window(char *fitsInname, int gp =0, int csiz = 2048) ;

	Window(string fitsInname, int gp =0, int csiz = 2048) ;

	// Window(const Window & othwin) ;

	~Window() ;

	void win2win(const int rotRight90, const bool flip) ;

	int sizeX() const ;

	int sizeY() const ;

	static bool isRef(int xin, int yin, int xchipsize, int ychipsize) ;

	bool isRef(int xin, int yin) const;

	void blankRef(int blank) ;

	void cooGeirsWinToSplicRot(int coo[2], const int rotRight90, const bool flip) ;

	static void fourPix(const string & tsiz, int coo[4],bool relwidth) ;

	string canonName() ;

	static string trimws(const string & instring) ;

	void rmFiles(bool verbose=false) ;

protected:

	void cooGeirsNoWin(int coo[2]) ;

	void cooToWin(int coo[2]) ;

	void detIdx(const int coo[2], int idx[2]) ;

	static void pixRot(int coo[2], int rotRight90, bool flip) ;

	void cooGeirsWinToSplic(int coo[2]) ;


	void cooGeirsWinToSplicRotWin(int coo[2], const int rotRight90, const bool flip) ;

private:
	void readPix(bool verbose);

	void init(bool verbose) ;

} ; /* Window */

#endif /* WINDOW_H */
