/*
* MPIA - GEIRS project
*
* "@(#) $Id$"
*
*/


#include <string>
#include <vector>
#include <fstream>
#include <limits>

#include "config.h"
#include "WindowSet.h"

// #include <boost/regex.hpp>

using namespace std ;
using namespace CCfits ;

/** 
* @param[in] fitsname The name of the FITS file to be read.
* @param[in] gp The width of the gap between individual chips in the mosaic in units of pixels.
* @param[in] csiz The dimension along each edge of each chip in pixels.
* @since 2012-10-23
* @author Richard J. Mathar
*/
WindowSet::WindowSet(const char *fitsIname, int gp, int csiz) : ws()
{
	/* construct a set of matching files (for GEIRS, these are the ????*_win.fits
	* or the single ????.fits file).
	*/
	vector<string> wfils = ifileset(fitsIname) ;


#ifdef DEBUG
	FITS::setVerboseMode(true) ;
#endif
	for(int i=0; i< wfils.size(); i++)
	{
		/* Create window using gp and csiz defaults, read its
		* pixels and location.
		*/
		// Window w(wfils[i],gp,csiz) ;
		Window * w =new Window(wfils[i],gp,csiz) ;

		/* add to the set of known windows
		*/
		ws.push_back(w) ;

#ifdef DEBUG
		cout << "...Input file " << wfils[i] << endl ;
#endif
		/* at this point w goes out of scope and the dtor of w was
		* called if this was just an instance and not a pointer...
		*/
	}

	/* create the default which will surely expand as new windows
	* are added to the group. So this will end up with the tight
	* frame around the components, once win2win() has been called.
	*/
	trimsize[0] = trimsize[2] = numeric_limits<int>::max() ;
	trimsize[1] = trimsize[3] = numeric_limits<int>::min() ;

	if ( wfils.size() == 0)
		cerr << __FILE__" Error: Cannot open/find files named " << fitsIname << endl ;
} /* WindowSet::WindowSet */

/** Remove the input files from the file system.
* This hazardous action removes the FITS files that have been used to
*  define the windows and their contents.
* @param[in] verbose If true, print each file's name  removed to stdout.
* @since 2012-11-20
* @author Richard J. Mathar
*/
void WindowSet::rmFiles(bool verbose)
{
	for(int i=0; i< ws.size(); i++)
		ws[i]->rmFiles(verbose) ;
} /* WindowSet::rmFiles */

/** Determine the list of all fits file names that match the base name.
* @param[in] fitsname The basename name of the FITS file to be read.
* @return The existing files of the format of the argument with suffix
*   .fits, and the existing files with suffices *_wini.fits, with i a
* set of consecutive increasing integer numbers.
*/
vector<string> WindowSet::ifileset(const char *fitsIname)
{
	/* group of all file names found
	*/
	vector<string> fnams ;

	/* first try to add the bare file name as given
	*/
	ostringstream tstNam ;
	tstNam << fitsIname ;
	// string tstNam = string(fitsIname) ;
	// ifstream inf(tstNam.c_str()) ;
	ifstream inf(tstNam.str().c_str()) ;
	if ( inf )
		fnams.push_back(tstNam.str()) ;

	/* second attempt: add the file name with suffix
	*/
	tstNam << ".fits" ;
	ifstream infExt(tstNam.str().c_str()) ;
	// tstNam = string(fitsIname)+ ".fits" ;
	// inf = ifstream(tstNam.c_str()) ;
	if ( infExt ) 
		fnams.push_back(tstNam.str()) ;

	/* third attempt: add the file names  with suffix _win*.fits
	*/
	for (int i=1 ; ; i++)
	{
		tstNam.seekp(0) ;
		tstNam  << fitsIname << "_win" << i << ".fits" ;
		// ifstream inf(tstNam.c_str()) ;
		ifstream inf(tstNam.str().c_str()) ;
		if ( inf )
			fnams.push_back(tstNam.str()) ;
		else
			/* leave loop if not found
			*/
			break;
	}
	return fnams ;
} /* ifileset */

/** Destructor.
*/
WindowSet::~WindowSet()
{
	/* this would not be needed since the dtor of vector() executes the individual
	* dtor's of the components.
	*/
	for (int w=0 ; w < ws.size() ; w++)
		delete ws[w] ;
} /* WindowSet::dtor */

/** Compute the window location after splicing and optional flip-rotation.
* @param[in] rotRight90 Number of right-90 rotations requested.
* @param[in] flip If true, perform antoher right-left flip around the y-coordinate.
*/
void WindowSet::win2win(const int rotRight90, const bool flip)
{
	/* figure out the trimsize[] of each individual subwindow,
	* and build the rectangular frame around this. Default values
	* have been set up in the ctor.
	*/
	for (int w=0 ; w < ws.size() ; w++)
	{
		// ws[w].win2win(rotRight90, flip) ;
		//trimsize[0] = min(trimsize[0],ws[w].trimsize[0]) ;
		//trimsize[1] = max(trimsize[1],ws[w].trimsize[1]) ;
		//trimsize[2] = min(trimsize[2],ws[w].trimsize[2]) ;
		//trimsize[3] = max(trimsize[3],ws[w].trimsize[3]) ;
		ws[w]->win2win(rotRight90, flip) ;
		trimsize[0] = min(trimsize[0],ws[w]->trimsize[0]) ;
		trimsize[1] = max(trimsize[1],ws[w]->trimsize[1]) ;
		trimsize[2] = min(trimsize[2],ws[w]->trimsize[2]) ;
		trimsize[3] = max(trimsize[3],ws[w]->trimsize[3]) ;
	}
} /* win2win */
