/** @file
* fitsImgRot is a C++ program which rotates a FITS image by a variable angle about a variable point.
*/

#include <unistd.h>
#include "PixImage.h"

/** Print a usage syntax message detailing the command line
* @param[in] argv0 The name of the main program.
* @since 2012-10-18
*/
void usage(char *argv0)
{
        cout << "Usage:" << endl ;
        cout << argv0 << " [-v] [-a angle/deg] [-x Npix] [-y Npix] [-s slice] fitsfilename outfitsfilename" << endl ;
} /* usage */


/*!***********************************************
* @brief The program rotates an image in the primary HDU of a FITS file.

* The program implements a drizzle algorithm to rotate the individual pixels
* of a FITS image by an arbitrary angle, and to reassemble the portion of light
* in the rotated image by binning along the two Cartesian axes.
*
* For a more flexible package see the montage library of http://montage.ipac.caltech.edu/ .
*
* @param argc The number of command line arguments (and options). Minimum of two.
* @param argv The program takes up to four options and exactly two arguments. In detail:
*  Option @code{-a} followed by a floating point number
*  specifies the angle in degrees. The original image will be
*  rotated counter clock wise by that angle around the origin (which is
*  defined here to be the lower left corner of the lower left pixel).
*  The default is 0.0, so basically all non-trivial uses of the program
*  will use this option.
*
*  Option @code{-x} followed by a floating point number
*  specifies the x-component of a translation of the image after the
*  rotation. The default is 0.0. Note that the final binning of the rotated image with another grid of
*  quadratic pixels implies that only values in the range 0. to +1. are relevant.
*  Any integer value of the shift is essentially absorbed by the extension of the
*  final grid by integer units.
*
*  Option @code{-y} followed by a floating point number
*  specifies the y-component of a translation of the image after the
*  rotation. The default is 0.0. Equivalent notes as for the x-component apply.

*  Option @code{-s} followed by a non-negative integer selects a slice
*  of the FITS cube of the primary HDU, supposed a FITS cube exists there.
*  Slices are enumerated starting with index 0.
*  The default is 0.

* @todo Copy the header cards of the original FITS file to the rotated image.
* @todo Allow more flexible representation in the rotated image than the float (bitpix=-32) type.
* @todo Modify WCS header cards if present in the original image.
* @todo Use the origin of the original WCS for the default pole of the rotation if available.
* @since 2013-07-03
* @author Richard J. Mathar
*/
int main(int argc, char *argv[])
{
	bool verbose = false ;
	double angle=0. ;
	double txy[2] ={0.,0.} ;

	/* first slice of a cube, if applicable
	*/
	int slice = 0 ;
	/* option character
	*/
	char oc ;
	while (  (oc=getopt(argc,argv,"va:x:y:s:")) != -1 )
	{
		switch(oc)
		{
		case 'v' :
			verbose = true ;
			break ;
		case 'a' :
			angle = atof(optarg) ;
			break ;
		case 'x' :
			txy[0] = atof(optarg) ;
			break ;
		case 'y' :
			txy[1] = atof(optarg) ;
			break ;
		case 's' :
			slice = atoi(optarg) ;
			break ;
		case '?' :
			cerr << "Invalid command line option " << optopt << endl ;
			break ;
		}
	}

        /* ensure that this is called with two file names
        */
        if ( optind +2 != argc )
        {
                usage(argv[0]) ;
                return 1 ;
        }

	/* convert angle from degrees to radians
	*/
	angle *= M_PI/180.0 ;

	FITS::setVerboseMode(true) ;
	/* todo: exit1 if not readable
	*/
	FITS fin(argv[optind]) ;

	PixImage origin(fin,slice) ;

	RotTrans2D rt(angle,txy[0],txy[1]) ;

#if 0
	PixImage img = origin.apply(rt) ;
	FITS * fout = img.toFits(argv[optind+1]) ;
#else
	PixImage *img = origin.apply(rt) ;

	FITS * fout = img->toFits(argv[optind+1]) ;
	delete img ;
#endif

	delete fout ;

	return 0 ;
} /* main */
