/** @file
* The program geirs2Panic is a postprocessor for FITS mosaic files that have been generated by GEIRS.
*/

/*
* MPIA - GEIRS project
*
* "@(#) $Id$"
*
*/

#include "config.h"
#include <unistd.h>
#include "GeirsPanic.h"

/** Print a usage syntax message detailing the command line
* @param[in] argv0 The name of the main program.
* @since 2012-10-18
*/
void usage(char *argv0, int gap,int chpsz, double north)
{
	cout << "Usage:" << endl ;
#if 0
	cout << argv0 << " [-c chipsz] [-g gapPx] [-D] [-A] [-t] [-f] [-N degr] fitsIn [fitsOut.fits]" << endl ;
#else
	cout << argv0 << " [-c chipsz] [-g gapPx] [-A] [-t] [-f] [-N degr] fitsIn [fitsOut.fits]" << endl ;
#endif
	cout << "Option -c is the size of the single chip in the mosaic. Default is " << chpsz << "." << endl ;
	cout << "Option -g is the gap in the mosaic. Default is " << gap << "." << endl ;
	cout << "Option -A triggers computation of ALT, AZ and PARANG" << "." << endl ;
#if 0
  	cout << "Option -D triggers computation of DATAMIN and DATAMAX." << endl 
#endif
	cout << "Option -t trims (removes) the 4 rows/columns of reference pixels." << endl ;
	cout << "Option -G adds OBSGEO keywords with WGS84 parameters." << endl ;
	cout << "Option -f flips the images (to swap E to the other side of N)." << endl ;
	cout << "Option -N indicates the direction of N in fitsIn.fits in degrees. 90 means up, 180 means left. Default is "
		<< north << endl ;
	cout << "Option -v increases the verbosity of the output." << endl ;
	cout << "The dangerous (!) option -r removes the (set of) fitsIn files when finished." << endl ;
	cout << "Argument fitsIn is either a single existing FITS file name or the stem (without the _win*) of a window set." << endl ;
	cout << "Argument fitsOut.fits is either a nonexisting FITS file name or generated automatically." << endl ;
} /* usage */

/** The program is a postprocessor for FITS files of CAHA/PANIC that have been generated by GEIRS.
*
* The main and mandatory arguments are an existing FITS file of the 2x2 mosaic
* in the glued representation as generated by GEIRS, and the name of a non-existing
* FITS file that will be generated by the program.
*
* The program supports
* - insertion of a cross of undefined values in the mosaic,
*   the definition of the North direction in the image and associated flip/rotation
*   by multiples of 90 degrees to align that direction as good as possible with
*   the up-direction in the image, plus insertion of the linear WCS scale transformation.
* - insertion of ALT, AZ and PARANG angles in the output header.
* - insertion of OBSGEO-X, OBSGEO-Y and OBSGEO-Z in the output header.
*
* Note that in a field of view stretching across 1 deg, the atmospheric
* dispersion R = tan(z)*(n-1) (in radians) differs across the field by
* delta(R) = delta(z)/cos^2(z) *(n-1), where n-1=2e-4 is roughly the
* index of refraction at the mountain altitude of 2700 m. Since delta(z)
* is roughly 4000 pixels, the product of both factors amounts to roughly 1 pixel across the field,
* but that expected linear distortion is not yet included in the WCS
* keywords of the header in the output.
*
* All of these tasks are actually covered by the instrument pipeline
* (Ibanez-Mengual et al, "The PANIC software system", SPIE 77402E, July 19 2012, doi:10.1117/12.856029)
* and do not need to be run on the mountain.
* 
* @param[in] argc The number of command line switches and arguments.
* @param[in] argv The vector of all command line arguments.
* The standard syntax is
* <code>
* geirs2Panic [-c chipsz] [-g gapPx] [-A] [-t] [-f] [-N degr] [-v] fitsIn [fitsOut.fits]
* </code>
* The options and their arguments are:
* - <code>-c cipsz</code> Sets the edge length of the single chip in the mosaic in units
*   of pixels. This defines the regular distance at which gaps appear in the mosaic.
*   The current default size is 2048. CAHA/LAICA users should set this to 4096.
* - <code>-g gapPx</code> This is the gap between any two chips in the mosaic, in units
*   of pixels. There is a default which is some value of the literature, currently 167.
*   CAHA/LAICA users should set this to 3207 (assuming a pixel scale of 0.225 arcsec/px and
*   2 times 100 arcsec overlap of the gap with the two adjacent chips).
* - <code>-A</code> Requests that the ALT, AZ and PARANG values are inserted into the
*   header (or replaced if they exist). This uses standard spherical trigonometry and
*   assumes that the latitude, hour angle and declination are all given in degrees
*   in the FITS header. The latitude may either be <code>TELLAT</code> (preferred is the
*   standard </code>OBSGEO-B</code>)
*   and the other two are taken from HA and DEC in the FITS header.
* - <code>-t</code> Replaces the 4 pixels around each chip of the mosaic by
*   the <code>BLANK</code> value. This is useful to avoid confusion of standard pipeline
*   code that does not know how to interpret these engineering pixels.
* - <code>-N degr</code> Indicates that North is that many degrees in counter-clock-wise
*   position from the horizontal axis in the input image. So a value near +90 would be
*   about ideal. This has two implications: The WCS sky transformation will have the
*   correct entries, and values more than 45 degrees away from 90 will trigger
*   a rotation by multiples of 90 degrees on output.
* - <code>-f</code> Indicates that East is on the wrong side of North in the input
*   data. This triggers that an additional flip (alongside any rotations) of the
*   images is done for the output. Note that this is associated with a flip of
*   North and additional changes of the WCS matrix.
* - <code>-G</code> Triggers that the <code>OBSGEO-X</code>
*     <code>OBSGEO-Y</code>
*     <code>OBSGEO-Z</code> triple is added to the header on output. This works
*   only if the three <code>TELLONG</code> <code>TELLAT</code> and <code>TELALT</code>
*   or their standard <code>OBSGEO-B</code>, <code>OBSGEO-L</code> and <code>OBSGEO-H</code>
*   values with are provided in the standard units in the input header.
* - <code>-v</code> generates more comments on standard output about what the
*   program is doing.
* - <code>fitsIn</code> is either (i) a single file with suffix <code>.fits</code> which contains
*   either a full frame or a subwindow of the mosaic, or (ii) a basis name for a set
*   of GEIRS window files without their <code>_win</code>i<code>.fits</code> parts.
*   Note that any of these windows may already stretch across more than one of the
*   chips of the mosaic, because GEIRS uses windowing in that manner.
* - <code>fitsOut.fits</code> is an optional file name with suffix <code>.fits</code> 
*   to be chosen for the output file. If this final argument is absent, the program
*   choses a file name which is the concatenation of the instrument name (keyword INSTRUME)
*   and date, keyword DATE-OBS (preferred) or DATE.
* @since 2012-10-23
*/
int main(int argc, char *argv[])
{
	bool doflip =false ;
	bool verbose =false ;
	bool doaltaz =false ;
	bool doTrim = false ;
	bool doGeo = false ;
	bool doMimax = false ;
	bool remov = false ;
	double north = 90. ;

	/* default gap is the gap in the PANIC detector mosaic
	*/
	int gap = 167 ;

	/* default chip size is Hawaii-II or HawiiII-2RD
	*/
	int chpsz = 2048 ;

	/* option character
	*/
	char oc ;

	while (  (oc=getopt(argc,argv,"g:c:AtDGfN:vr")) != -1 )
	{
		switch(oc)
		{
		case 'v' :
			verbose = true ;
			break ;
		case 'r' :
			remov = true ;
			break ;
		case 'g' :
			gap =atoi(optarg) ;
			break ;
		case 'c' :
			chpsz =atoi(optarg) ;
			break ;
		case 'A' :
			doaltaz = true ;
			break ;
		case 'f' :
			doflip = true ;
			break ;
		case 't' :
			doTrim = true ;
			break ;
		case 'G' :
			doGeo = true ;
			break ;
#if 0
		/* permamently disabled because we use the "fast" deletion if blanking the edge
		* pixels and these would (falsely) be included in the minmax determination of kept in.
		*/
		case 'D' :
			doMimax = true ;
			break ;
#endif
		case 'N' :
			north =atof(optarg) ;
			break ;
		case '?' :
			cerr << "Invalid command line option " << optopt << endl ;
			break ;
		}
	}

	/* ensure that this is called with two file names
	*/
	if ( optind >= argc )
	{
		usage(argv[0],gap,chpsz,north) ;
		return 1 ;
	}

	GeirsPanic fedt(argv[optind], (optind+1 < argc) ? argv[optind+1] : "" ,gap,chpsz) ;
	fedt.exec(doTrim, doMimax, north, doflip, doaltaz, doGeo, verbose,remov) ;

	return 0 ;
} /* main */
